ASSOCIATION_LN = "Соединение по логическому имени"
ARBITRATOR = "Арбитр"
PSTN_MODEM_CONFIGURATION = "PSTN настройки модема"
MODEM_CONFIGURATION = "Настройки модема"
SECURITY_SETUP = "Параметры безопасности"
ACTIVITY_CALENDAR = "Календарь активности"
CLIENT_SETUP = "Настройки клиента"
CLOCK = "Часы"
DATA = "Данные"
DISCONNECT_CONTROL = "Управление отключением"
EXTENDED_REGISTER = "Расширенный регистр"
DEMAND_REGISTER = "Регистр усреднения"
GPRS_MODEM_SETUP = "GPRS настройки модема"
GSM_DIAGNOSTIC = "GSM диагностика"
IEC_HDLC_SETUP = "IEC HDLC настройки"
IEC_LOCAL_PORT_SETUP = "IEC настройки локального порта"
IMAGE_TRANSFER = "Передача двоичных блоков"
IPV4_SETUP = "IPv4 настройки"
LIMITER = "Ограничитель"
PROFILE_GENERIC = "Профиль универсальный"
PUSH_SETUP = "Настройки инициативного выхода"
REGISTER = "Регистр"
REGISTER_MONITOR = "Регистр контроля"
SCHEDULE = "Расписание"
SCRIPT_TABLE = "Таблица сценариев"
SINGLE_ACTION_SCHEDULE = "Расписание одного действия"
SPECIAL_DAYS_TABLE = "Таблица особых дней"
TCP_UDP_SETUP = "TCP UDP настройки"
