MANUAL = "Ручное (по кнопке)"
LOCAL_1 = "Локальное. Aктивная мощность"
LOCAL_2 = "Локальное. Ток"
LOCAL_3 = "Локальное. Напряжение"
LOCAL_4 = "Локальное. Магнитное поле"
LOCAL_5 = "Локальное. Небаланс токов"
LOCAL_6 = "Локальное. Температура"
LOCAL_7 = "Локальное. Регистр монитор"
